﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  [Serializable]
  [StorableClass]
  public abstract class DoCountExpression : LoopExpression {

    protected DoCountExpression() { }
    protected DoCountExpression(LoopState state) : base(state) { }
    [StorableConstructor]
    protected DoCountExpression(bool deserializing) : base(deserializing) { }
    protected override bool HasInsufficientArguments(IInternalPushInterpreter interpreter, IPushStack<Expression> sourceStack) {
      return interpreter.IntegerStack.Count < 1 ||
             sourceStack.Count == 0 ||
             interpreter.IntegerStack.Top <= 0;
    }

    protected override LoopState InitState(IInternalPushInterpreter interpreter, IPushStack<Expression> sourceStack) {
      var state = LoopState.Create(interpreter.PoolContainer.LoopStatePool,
        body: sourceStack.Pop(),
        currentIndex: 1,
        destinationIndex: interpreter.IntegerStack.Top,
        incrementor: 1);

      interpreter.IntegerStack.SetTop(0);

      return state;
    }
  }

  /// <summary>
  ///     An iteration instruction that performs a loop (the body of which is taken from the CODE stack) the number of times
  ///     indicated by
  ///     the INTEGER argument, pushing an index (which runs from zero to one less than the number of iterations) onto the
  ///     INTEGER stack
  ///     prior to each execution of the loop body. This should be implemented as a macro that expands into a call to
  ///     CODE.DO*RANGE.
  ///     CODE.DO*COUNT takes a single INTEGER argument (the number of times that the loop will be executed) and a single
  ///     CODE argument
  ///     (the body of the loop). If the provided INTEGER argument is negative or zero then this becomes a NOOP. Otherwise it
  ///     expands into:
  ///     ( 0 '1 - IntegerArg>' CODE.QUOTE 'CodeArg' CODE.DO*RANGE )
  /// </summary>
  [Serializable]
  [StorableClass]
  [PushExpression(StackTypes.Code, "CODE.DO*COUNT", StackTypes.Integer)]
  public class CodeDoCountExpression : DoCountExpression {

    public CodeDoCountExpression() { }
    public CodeDoCountExpression(LoopState state) : base(state) { }
    [StorableConstructor]
    protected CodeDoCountExpression(bool deserializing) : base(deserializing) { }
    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return State.Body == null && HasInsufficientArguments(interpreter, interpreter.CodeStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.CodeStack);
    }

    protected override LoopExpression Clone(LoopState state, IInternalPushInterpreter interpreter) {
      var expression = interpreter.PoolContainer.GetStatefulExpression<CodeDoCountExpression>();
      expression.State = state;

      return expression;
    }
  }

  /// <summary>
  ///     An iteration instruction that performs a loop (the body of which is taken from the EXEC stack) the number of times
  ///     indicated
  ///     by the INTEGER argument, pushing an index (which runs from zero to one less than the number of iterations) onto the
  ///     INTEGER
  ///     stack prior to each execution of the loop body. This is similar to CODE.DO*COUNT except that it takes its code
  ///     argument from
  ///     the EXEC stack. This should be implemented as a macro that expands into a call to EXEC.DO*RANGE. EXEC.DO*COUNT
  ///     takes a single
  ///     INTEGER argument (the number of times that the loop will be executed) and a single EXEC argument (the body of the
  ///     loop). If
  ///     the provided INTEGER argument is negative or zero then this becomes a NOOP. Otherwise it expands into:
  ///     ( 0 '1 - IntegerArg' EXEC.DO* RANGE'ExecArg' )
  /// </summary>
  [Serializable]
  [StorableClass]
  [PushExpression(StackTypes.Exec, "EXEC.DO*COUNT", StackTypes.Integer)]
  public class ExecDoCountExpression : DoCountExpression {
    public ExecDoCountExpression() { }
    public ExecDoCountExpression(LoopState state) : base(state) { }
    [StorableConstructor]
    protected ExecDoCountExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.ExecStack);
    }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return State.Body == null && HasInsufficientArguments(interpreter, interpreter.ExecStack);
    }

    protected override LoopExpression Clone(LoopState state, IInternalPushInterpreter interpreter) {
      var expression = interpreter.PoolContainer.GetStatefulExpression<ExecDoCountExpression>();
      expression.State = state;

      return expression;
    }
  }
}