﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using System.Linq;

  using Attributes;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

  using Interpreter;
  using Stack;

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.FROMINTEGER", StackTypes.Integer)]
  public class StringFromIntegerExpression : StatelessExpression {
    public StringFromIntegerExpression() { }
    [StorableConstructor]
    protected StringFromIntegerExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = interpreter.IntegerStack.Pop();
      interpreter.StringStack.Push(value.ToString());
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.FROMFLOAT", StackTypes.Float)]
  public class StringFromFloatExpression : StatelessExpression {
    public StringFromFloatExpression() { }
    [StorableConstructor]
    protected StringFromFloatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = interpreter.FloatStack.Pop();
      interpreter.StringStack.Push(value.ToString());
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.FROMBOOLEAN", StackTypes.Boolean)]
  public class StringFromBooleanExpression : StatelessExpression {
    public StringFromBooleanExpression() { }
    [StorableConstructor]
    protected StringFromBooleanExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = interpreter.BooleanStack.Pop();
      interpreter.StringStack.Push(value.ToString());
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.FROMCHAR", StackTypes.Char)]
  public class StringFromCharExpression : StatelessExpression {
    public StringFromCharExpression() { }
    [StorableConstructor]
    protected StringFromCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = interpreter.CharStack.Pop();
      interpreter.StringStack.Push(value.ToString());
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.CONCAT")]
  public class StringConcatExpression : StatelessExpression {
    public StringConcatExpression() { }
    [StorableConstructor]
    protected StringConcatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.Count < 2 ||
             interpreter.StringStack.Top.Length + interpreter.StringStack[1].Length >= interpreter.Configuration.MaxStringLength;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Pop();
      interpreter.StringStack.SetTop(str + interpreter.StringStack.Top);
    }
  }

  /// <summary>
  /// Conj char onto string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.CONJCHAR", StackTypes.Char)]
  public class StringConjCharExpression : StatelessExpression {
    public StringConjCharExpression() { }
    [StorableConstructor]
    protected StringConjCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.IsEmpty ||
             interpreter.StringStack.Top.Length + 1 >= interpreter.Configuration.MaxStringLength;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var c = interpreter.CharStack.Pop();
      interpreter.StringStack.SetTop(interpreter.StringStack.Top + c);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.TAKE", StackTypes.Integer)]
  public class StringTakeExpression : StatelessExpression {
    public StringTakeExpression() { }
    [StorableConstructor]
    protected StringTakeExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.IntegerStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = interpreter.IntegerStack.Pop();
      var str = interpreter.StringStack.Top;

      if (str.Length == 0)
        return;

      value = Math.Abs(value) % str.Length;
      interpreter.StringStack.SetTop(str.Substring(0, (int)value));
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.SUBSTRING", StackTypes.Integer)]
  public class StringSubstringExpression : StatelessExpression {
    public StringSubstringExpression() { }
    [StorableConstructor]
    protected StringSubstringExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.IntegerStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      var firstValue = interpreter.IntegerStack[1];
      var secondValue = interpreter.IntegerStack.Top;
      interpreter.IntegerStack.Remove(2);

      var first = Math.Min(str.Length - 1, Math.Max(firstValue, 0));
      var second = Math.Min(str.Length - 1, Math.Max(secondValue, first));
      var length = second - first;

      if (length > 0)
        interpreter.StringStack.SetTop(str.Substring((int)first, (int)length));
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.FIRST")]
  public class StringFirstExpression : StatelessExpression {
    public StringFirstExpression() { }
    [StorableConstructor]
    protected StringFirstExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.StringStack.Top.Length == 0;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      interpreter.StringStack.SetTop(interpreter.StringStack.Top[0].ToString());
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.LAST")]
  public class StringLastExpression : StatelessExpression {
    public StringLastExpression() { }
    [StorableConstructor]
    protected StringLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.StringStack.Top.Length == 0;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      var c = str[str.Length - 1].ToString();
      interpreter.StringStack.SetTop(c);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.NTH", StackTypes.Integer)]
  public class StringNthExpression : StatelessExpression {
    public StringNthExpression() { }
    [StorableConstructor]
    protected StringNthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.IntegerStack.IsEmpty ||
             interpreter.StringStack.Top.Length == 0;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      var index = str.Length == 1 ? 0 : (int)Math.Abs(interpreter.IntegerStack.Pop()) % str.Length;
      var c = str[index].ToString();
      interpreter.StringStack.SetTop(c);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REST")]
  public class StringRestExpression : StatelessExpression {
    public StringRestExpression() { }
    [StorableConstructor]
    protected StringRestExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.StringStack.Top.Length == 0;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      interpreter.StringStack.SetTop(str.Length == 1 ? string.Empty : str.Substring(1, str.Length - 1));
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.BUTLAST")]
  public class StringButLastExpression : StatelessExpression {
    public StringButLastExpression() { }
    [StorableConstructor]
    protected StringButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.StringStack.Top.Length == 0;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      interpreter.StringStack.SetTop(str.Length == 1 ? string.Empty : str.Substring(0, str.Length - 1));
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.LENGTH", StackTypes.Integer)]
  public class StringLengthExpression : StatelessExpression {
    public StringLengthExpression() { }
    [StorableConstructor]
    protected StringLengthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Pop();
      interpreter.IntegerStack.Push(str.Length);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REVERSE")]
  public class StringReverseExpression : StatelessExpression {
    public StringReverseExpression() { }
    [StorableConstructor]
    protected StringReverseExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      interpreter.StringStack.SetTop(interpreter.StringStack.Top.Reverse().ToString());
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.PARSETOCHARS")]
  public class StringParseToCharsExpression : StatelessExpression {
    public StringParseToCharsExpression() { }
    [StorableConstructor]
    protected StringParseToCharsExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.StringStack.Top.Length == 0) {
        interpreter.StringStack.Pop();
        return;
      }

      var str = interpreter.StringStack.Top;
      interpreter.StringStack.SetTop(str[0].ToString());

      if (str.Length > 1) {
        var chars = new string[str.Length - 1];
        for (var i = 0; i < str.Length - 1; i++) {
          chars[i] = str[i + 1].ToString();
        }

        interpreter.StringStack.Push(chars);
      }
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.SPLIT")]
  public class StringSplitExpression : StatelessExpression {
    public StringSplitExpression() { }
    [StorableConstructor]
    protected StringSplitExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.StringStack.Top.Trim().Split().Length <= 1;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var words = interpreter.StringStack.Top.Trim().Split();

      interpreter.StringStack.SetTop(words[0]);
      interpreter.StringStack.Push(words, 1);
    }
  }

  /// <summary>
  /// True if top string is empty
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.EMPTYSTRING", StackTypes.Boolean)]
  public class StringEmptyStringExpression : StatelessExpression {
    public StringEmptyStringExpression() { }
    [StorableConstructor]
    protected StringEmptyStringExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Pop();
      interpreter.BooleanStack.Push(str.Length == 0);
    }
  }

  /// <summary>
  /// True if top string is a substring of second string; false otherwise
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.CONTAINS", StackTypes.Boolean)]
  public class StringContainsExpression : StatelessExpression {
    public StringContainsExpression() { }
    [StorableConstructor]
    protected StringContainsExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.StringStack[1];
      var second = interpreter.StringStack.Top;
      interpreter.StringStack.Remove(2);

      interpreter.BooleanStack.Push(first.IndexOf(second, StringComparison.Ordinal) >= 0);
    }
  }

  /// <summary>
  /// True if the top char is in the top string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.CONTAINSCHAR", StackTypes.Boolean | StackTypes.Char)]
  public class StringContainsCharExpression : StatelessExpression {
    public StringContainsCharExpression() { }
    [StorableConstructor]
    protected StringContainsCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Pop();
      var c = interpreter.CharStack.Pop();

      interpreter.BooleanStack.Push(str.IndexOf(c) >= 0);
    }
  }

  /// <summary>
  /// Puts on the integer stack the index of the top char in the top string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.INDEXOFCHAR", StackTypes.Integer | StackTypes.Char)]
  public class StringIndexOfCharExpression : StatelessExpression {
    public StringIndexOfCharExpression() { }
    [StorableConstructor]
    protected StringIndexOfCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Pop();
      var c = interpreter.CharStack.Pop();

      interpreter.IntegerStack.Push(str.IndexOf(c));
    }
  }

  /// <summary>
  /// The number of times the top char is in the top string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.OCCURENCESOFCHAR", StackTypes.Integer | StackTypes.Char)]
  public class StringOccurrencesOfCharExpression : StatelessExpression {
    public StringOccurrencesOfCharExpression() { }
    [StorableConstructor]
    protected StringOccurrencesOfCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Pop();
      var c = interpreter.CharStack.Pop();

      var count = 0;
      for (var i = 0; i < str.Length; i++)
        if (str[i] == c) count++;

      interpreter.IntegerStack.Push(count);
    }
  }

  /// <summary>
  /// In third string on stack, replaces second string with first string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REPLACE")]
  public class StringReplaceExpression : StatelessExpression {
    public StringReplaceExpression() { }
    [StorableConstructor]
    protected StringReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.StringStack[1];
      var second = interpreter.StringStack.Top;
      interpreter.StringStack.Remove(2);

      if (first.Length == 0 || second.Length == 0)
        interpreter.StringStack.SetTop(first);
      else {
        var result = first.Replace(second, interpreter.StringStack.Top);
        interpreter.StringStack.SetTop(result);
      }
    }
  }

  /// <summary>
  /// In third string on stack, replaces first occurence of second string with first string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REPLACEFIRST")]
  public class StringReplaceLastExpression : StatelessExpression {
    public StringReplaceLastExpression() { }
    [StorableConstructor]
    protected StringReplaceLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.StringStack[0];
      var second = interpreter.StringStack[1];
      var third = interpreter.StringStack[2];
      interpreter.StringStack.Remove(2);

      var result = ReplaceFirst(third, second, first);
      interpreter.StringStack.SetTop(result);
    }

    private static string ReplaceFirst(string text, string search, string replace) {
      int pos = text.IndexOf(search);
      if (pos < 0) {
        return text;
      }
      return text.Substring(0, pos) + replace + text.Substring(pos + search.Length);
    }
  }

  /// <summary>
  /// In top string on stack, replaces all occurences of second char with first char
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REPLACECHAR", StackTypes.Char)]
  public class StringReplaceCharExpression : StatelessExpression {
    public StringReplaceCharExpression() { }
    [StorableConstructor]
    protected StringReplaceCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.CharStack[1];
      var second = interpreter.CharStack.Top;
      interpreter.CharStack.Remove(2);

      var result = interpreter.StringStack.Top.Replace(first, second);
      interpreter.StringStack.SetTop(result);
    }
  }

  /// <summary>
  /// In top string on stack, replaces first occurence of second char with first char
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REPLACEFIRSTCHAR", StackTypes.Char)]
  public class StringReplaceLastCharExpression : StatelessExpression {
    public StringReplaceLastCharExpression() { }
    [StorableConstructor]
    protected StringReplaceLastCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      var first = interpreter.CharStack[1];
      var second = interpreter.CharStack.Top;
      interpreter.CharStack.Remove(2);

      var pos = str.IndexOf(first);

      if (pos < 0)
        return;

      var result = str.Substring(0, pos) +
                   second +
                   str.Substring(Math.Min(pos + 2, str.Length - 1));

      interpreter.StringStack.SetTop(result);
    }
  }

  /// <summary>
  /// In top string on stack, remove all occurences of char
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.REMOVECHAR", StackTypes.Char)]
  public class StringRemoveCharExpression : StatelessExpression {
    public StringRemoveCharExpression() { }
    [StorableConstructor]
    protected StringRemoveCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var c = interpreter.CharStack.Pop();
      var result = interpreter.StringStack.Top.Trim(c);
      interpreter.StringStack.SetTop(result);
    }
  }

  /// <summary>
  /// Returns a function that sets char at index in string
  /// </summary>
  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.SETCHAR", StackTypes.Char | StackTypes.Integer)]
  public class StringSetCharExpression : StatelessExpression {
    public StringSetCharExpression() { }
    [StorableConstructor]
    protected StringSetCharExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.CharStack.IsEmpty ||
             interpreter.IntegerStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var str = interpreter.StringStack.Top;
      var i = (int)interpreter.IntegerStack.Pop();
      var c = interpreter.CharStack.Pop();

      if (str.Length == 0) {
        interpreter.StringStack.Pop();
        return;
      }

      var pos = str.Length == 1 ? 0 : Math.Abs(i) % (str.Length - 1);
      var result = str.Substring(0, pos) + c + str.Substring(Math.Min(pos + 2, str.Length - 1));

      interpreter.StringStack.SetTop(result);
    }
  }
}
