﻿using System.Collections.Generic;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Gets the nth item from the type stack
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorNthExpression<T> : StatelessExpression {
    protected VectorNthExpression() { }
    [StorableConstructor]
    protected VectorNthExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IInternalPushInterpreter interpreter, IPushStack<List<T>> vectorStack) {
      return vectorStack.IsEmpty || vectorStack.Top.Count == 0 || interpreter.IntegerStack.IsEmpty;
    }

    protected void Eval(
      IInternalPushInterpreter interpreter,
      IPushStack<List<T>> vectorStack,
      IPushStack<T> literalStack) {
      var vector = vectorStack.Pop();
      var n = (int)interpreter.IntegerStack.Pop() % vector.Count;

      if (n < 0) n = n + vector.Count;

      literalStack.Push(vector[n]);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].NTH", StackTypes.Integer)]
  public class IntegerVectorNthExpression : VectorNthExpression<long> {
    public IntegerVectorNthExpression() { }
    [StorableConstructor]
    protected IntegerVectorNthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.IntegerVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].NTH", StackTypes.Float | StackTypes.Integer)]
  public class FloatVectorNthExpression : VectorNthExpression<double> {
    public FloatVectorNthExpression() { }
    [StorableConstructor]
    protected FloatVectorNthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.FloatVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].NTH", StackTypes.Boolean | StackTypes.Integer)]
  public class BooleanVectorNthExpression : VectorNthExpression<bool> {
    public BooleanVectorNthExpression() { }
    [StorableConstructor]
    protected BooleanVectorNthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.BooleanVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "STRING[].NTH", StackTypes.String | StackTypes.Integer)]
  public class StringVectorNthExpression : VectorNthExpression<string> {
    public StringVectorNthExpression() { }
    [StorableConstructor]
    protected StringVectorNthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.StringVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
