﻿using System.Collections.Generic;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Replaces all occurrences of the second lit-type item with the first lit-type item in the top type vector.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorReplaceExpression<T> : StatelessExpression {
    protected VectorReplaceExpression() { }
    [StorableConstructor]
    protected VectorReplaceExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<List<T>> vectorStack, IPushStack<T> literalStack) {
      return vectorStack.IsEmpty || literalStack.Count < 2;
    }

    protected void Eval(IPushStack<List<T>> vectorStack, IPushStack<T> literalStack) {
      var vector = vectorStack.Top;
      var first = literalStack.Top;
      var second = literalStack[1];
      literalStack.Remove(2);

      for (var i = 0; i < vector.Count; i++) {
        if (vector[i].Equals(second))
          vector[i] = first;
      }
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].REPLACE", StackTypes.Integer)]
  public class IntegerVectorReplaceExpression : VectorReplaceExpression<long> {
    public IntegerVectorReplaceExpression() { }
    [StorableConstructor]
    protected IntegerVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].REPLACE", StackTypes.Float)]
  public class FloatVectorReplaceExpression : VectorReplaceExpression<double> {
    public FloatVectorReplaceExpression() { }
    [StorableConstructor]
    protected FloatVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack, interpreter.FloatStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].REPLACE", StackTypes.Boolean)]
  public class BooleanVectorReplaceExpression : VectorReplaceExpression<bool> {
    public BooleanVectorReplaceExpression() { }
    [StorableConstructor]
    protected BooleanVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "STRING[].REPLACE", StackTypes.String)]
  public class StringVectorReplaceExpression : VectorReplaceExpression<string> {
    public StringVectorReplaceExpression() { }
    [StorableConstructor]
    protected StringVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack, interpreter.StringStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
